using System;
using System.Reflection;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Drawing.Text;
using System.Windows.Forms;
using System.Collections;

[assembly:AssemblyKeyFileAttribute("../fontlib.snk")]

namespace Ishiboo {
    public class FontLib {
	private unsafe static void dropshadow(BitmapData bmp, int sz) {
	    byte *line2 = (byte*)bmp.Scan0 + (bmp.Stride * (bmp.Height-1)) + (4 * (bmp.Width-1));
	    byte *line = line2 - (bmp.Stride * sz) - (4 * sz);
	    int y = 0;
	    int h = bmp.Height - sz;
	    int w = bmp.Width - sz;

	    while (y < h) {
		byte *p = line;
		byte *p2 = line2;
		int x = 0;
		while (x < w) {
		    if (p[3] == 255) {
			// solid
			if (p2[3] == 0) { // if the destination has nothing
			    p2[0] = 0;
			    p2[1] = 0;
			    p2[2] = 0;
			    p2[3] = 255;
			} else if (p2[3] != 255) {
			    p2[0] = (byte)(p2[0] * p2[3] / 255);
			    p2[1] = (byte)(p2[1] * p2[3] / 255);
			    p2[2] = (byte)(p2[2] * p2[3] / 255);
			    p2[3] = 255;
			}
		    } else if (p[3] != 0) {
			// antialised edges
			if (p2[3] == 0) { // if the destination has nothing
			    p2[0] = 0;
			    p2[1] = 0;
			    p2[2] = 0;
			    p2[3] = p[3];
			}
		    }
		    p -= 4;
		    p2 -= 4;
		    x++;
		}
		line -= bmp.Stride;
		line2 -= bmp.Stride;
		y++;
	    }
	}

	private unsafe static void addbevel(BitmapData bmp, int sz) {
	    byte *line = (byte*)bmp.Scan0;
	    byte *line2 = line + (bmp.Stride * sz) + (4 * sz);
	    int y = 0;
	    int h = bmp.Height - sz;
	    int w = bmp.Width - sz;

	    while (y < h) {
		byte *p = line;
		byte *p2 = line2;
		int x = 0;
		while (x < w) {
		    int d1 = p2[3] - p[3];
		    if (d1 > 0) {
			p2[0] += (byte)((255 - p2[0]) * d1 / 255 / 5);
			p2[1] += (byte)((255 - p2[1]) * d1 / 255 / 5);
			p2[2] += (byte)((255 - p2[2]) * d1 / 255 / 5);
		    } else {
			int d2 = p[3] - p2[3];
			if (d2 > 0) {
			    p[0] -= (byte)(p[0] * d2 / 255 / 5);
			    p[1] -= (byte)(p[1] * d2 / 255 / 5);
			    p[2] -= (byte)(p[2] * d2 / 255 / 5);
			}
		    }
		    p += 4;
		    p2 += 4;
		    x++;
		}
		line += bmp.Stride;
		line2 += bmp.Stride;
		y++;
	    }
	}

	/// <summary>
	/// Draws a string in a provided font with a given background image
	/// that can be either scaled or tiled. A bevel can also be placed on
	/// the rendered characters. The font will be antialiased.
	/// </summary>
	public static Image DrawString(string s, Font f, Image bg,
				       bool scale_width,
				       bool scale_height,
				       bool bevel,
				       int dropshadowsz)
	{
	    if (!(bg is Bitmap))
		throw new Exception("background image is not a System.Drawing.Bitmap");


	    int w, h;
	    {// figure out width and height
		Graphics g = Graphics.FromImage(bg);
		SizeF sz = g.MeasureString(s, f);
		w = (int)sz.Width;
		h = (int)sz.Height;
	    }

	    if (scale_width || scale_height) {
		int ww, hh;
		if (scale_width) ww = w;
		else ww = bg.Width;
		if (scale_height) hh = h;
		else hh = bg.Height;
		bg = new Bitmap(bg, ww, hh);
	    }

	    Bitmap bmp = new Bitmap(w+dropshadowsz, h+dropshadowsz);
	    Graphics gfx = Graphics.FromImage(bmp);
	    gfx.TextRenderingHint = TextRenderingHint.AntiAlias;
	    gfx.DrawString(s, f, new TextureBrush(bg),
			   new PointF(0.0F, 0.0F));

	    BitmapData bmpdata =
		bmp.LockBits(new Rectangle(0, 0, bmp.Width, bmp.Height),
			     ImageLockMode.ReadWrite,
			     PixelFormat.Format32bppArgb);

	    if (bevel)
		addbevel(bmpdata, 2);

	    if (dropshadowsz != 0)
		dropshadow(bmpdata, dropshadowsz);

	    bmp.UnlockBits(bmpdata);

	    return bmp;
	}

	/// <summary>
	/// Draws a string in a provided font in a given color. A bevel can
	/// also be placed on the rendered characters. The font will be
	/// antialiased.
	/// </summary>
	public static Image DrawString(string s, Font f, Color solid,
				       bool bevel, int dropshadowsz)
	{
	    int w, h;
	    {// figure out width and height
		Image i = new Bitmap(1,1);
		Graphics g = Graphics.FromImage(i);
		SizeF sz = g.MeasureString(s, f);
		w = (int)sz.Width;
		h = (int)sz.Height;
	    }

	    Bitmap bmp = new Bitmap(w+dropshadowsz, h+dropshadowsz);

	    Graphics gfx = Graphics.FromImage(bmp);
	    gfx.TextRenderingHint = TextRenderingHint.AntiAlias;
	    gfx.DrawString(s, f, new SolidBrush(solid),
			   new PointF(0.0F, 0.0F));

	    BitmapData bmpdata =
		bmp.LockBits(new Rectangle(0, 0, bmp.Width, bmp.Height),
			     ImageLockMode.ReadWrite,
			     PixelFormat.Format32bppArgb);

	    if (bevel)
		addbevel(bmpdata, 2);

	    if (dropshadowsz != 0)
		dropshadow(bmpdata, dropshadowsz);

	    bmp.UnlockBits(bmpdata);

	    return bmp;
	}
    }
}
